(* $Id: builtins.ml,v 1.16 1997/04/14 02:25:33 garrigue Exp $ *)

#open"misc";;
#open"common";;
#open"types";;
#open"define";;
#open"predef";;

(* exception handlers *)

let handlers = ref [] ;;

let handle exn =
  exists (fun handler -> handler exn) !handlers
;;

(* user exception *)

exception User of string;;

let minimal_fail s = raise (User (string_of_array s)) ;;

handlers :=
  [fun exn ->
    let msg =
      match exn with
	User s -> "Exception : " ^ s
      | compile__Match_error s ->
	  if s <> "" then
	    "Match failure in " ^ s
	  else
	    "Match failure in toplevel input"
      | Not_found ->
	  "Internal : Not_found"
      | Invalid_argument "vect_item" ->
	  "Error : array access out of bounds"
      | Invalid_argument "vect_assign" ->
	  "Error : array assignment out of bounds"
      | Invalid_argument s ->
	  "Invalid argument for " ^ s
      | Division_by_zero ->
	  "Division by zero"
      | Failure s ->
	  "Failure : " ^ s
      | exn -> ""
    in
    if msg = "" then false else begin
      prerr_endline ("> " ^ msg ^ ".");
      true
    end]
;;

(* custom primitives *)

(* careful functions *)

let maxint = (-1) lsr 1 ;;
let minint = maxint + 1 ;;

let float_maxint = float_of_int maxint
and float_minint = float_of_int minint ;;

let truncate x =
  if x > float_maxint then maxint else
  if x < float_minint then minint else
  int_of_float x

and sqrt x = if x < 0. then raise (Invalid_argument "sqrt") else sqrt x

and log x = if x < 0. then raise (Invalid_argument "log") else log x

and asin x =
  if x < -1. or x > 1. then raise (Invalid_argument "asin") else asin x

and acos x =
  if x < -1. or x > 1. then raise (Invalid_argument "acos") else acos x

and div_float x y =
  if y = 0. then raise Division_by_zero else x /. y
;;

(* tracing *)

#open"obj";;
#open"format";;

let rec trace_obj s ty =
  match unify__repr ty with
    Tarrow (ty1, ty2) ->
      let tracing = trace_obj (s ^ " *") ty2 in
      (fun obj -> repr (fun arg ->
	open_hovbox 2;
	print_string (s ^ " <=");
	print_space ();
	printer__print_value ty1 arg;
	close_box ();
	print_newline ();
	let res = magic_obj obj arg in
	open_hovbox 2;
	print_string (s ^ " * =");
	print_space ();
	printer__print_value ty2 res;
	close_box ();
	print_newline ();
	tracing res))
  | _ -> (fun obj -> obj)
;;

let tracing_obj s ty_args ty_arg ty_res =
  let tracing = trace_obj (s ^ " ..") ty_res in
  (fun args obj -> repr(fun arg ->
    open_hovbox 2;
    print_string (s ^ " <=");
    do_list2
      (fun arg ty -> print_space(); printer__print_value ty arg)
      (rev (arg::args)) (rev (ty_arg::ty_args));
    close_box ();
    print_newline ();
    let res = magic_obj obj arg in
    open_hovbox 2;
    print_string (s ^ " =>");
    print_space ();
    printer__print_value ty_res res;
    close_box ();
    print_newline ();
    tracing res))
;;

let rec tracen_obj s n ty_args ty_res =
  match unify__repr ty_res with
    Tarrow (ty_arg, ty_res) ->
      if n = 1 or
	(match unify__repr ty_res with Tarrow _ -> false | _ -> true)
      then tracing_obj s ty_args ty_arg ty_res
      else begin
	let tracing = tracen_obj s (n-1) (ty_arg::ty_args) ty_res in
	(fun args obj -> repr (fun arg ->
	  let partial = magic_obj obj arg in
	  tracing (arg::args) partial))
      end
  | _ -> failwith "builtins__tracen_obj"
;;
  

#close"format";;

let trace_table = ref (map__empty compare_id : (ident,obj) map__t) ;;

let trace n s =
  let s = string_of_array s in
  try
    let info = map__find s !values
    and id = assoc s !global_idents in
    let obj_ref = map__find id !global_env in
    let old_obj =
      try map__find id !trace_table
      with Not_found -> !obj_ref in
    let is_function =
      match unify__repr info.vi_type with Tarrow _ -> true | _ -> false in
    if not is_function then
      prerr_endline (s ^ " is not a function, you cannot trace it.")
    else
      let traced_obj =
	if n <= 1 then trace_obj s info.vi_type old_obj
	else tracen_obj s n [] info.vi_type [] old_obj in
      trace_table := map__add id old_obj !trace_table;
      obj_ref := traced_obj;
      print_endline ("Function " ^ s ^ " is now traced.")
  with
    Not_found ->
      prerr_endline ("Unknown identifier " ^ s ^ ".")
;;

let untrace s =
  let s = string_of_array s in
  try
    let id = assoc s !global_idents in
    let obj_ref = map__find id !global_env
    and old_obj = map__find id !trace_table in
    obj_ref := old_obj;
    trace_table := map__remove id !trace_table;
    print_endline ("Function " ^ s ^ " will not be traced anymore.")
  with Not_found ->
    prerr_endline ("Function " ^ s ^ " is not traced.")
;;

let untrace_all () =
  map__iter
    (fun id old_obj ->
      let obj_ref = map__find id !global_env in
      obj_ref := old_obj)
    !trace_table;
  trace_table := map__empty compare_id;
  print_endline "All functions are no longer traced."
;;


(* definitions *)

let int = Tconstr(id_int,[])
and char = Tconstr(id_char,[])
and float = Tconstr(id_float,[])
and bool = Tconstr(id_bool,[])
and unit = Ttuple[]
and a = Tvar{link=None;level= -1}
and b = Tvar{link=None;level= -1}
and arr t t' = Tarrow(t,t')
and arr2 t1 t2 t = Tarrow(t1,Tarrow(t2,t))
and list a = Tconstr(id_list,[a])
and array a = Tconstr(id_array,[a])
;;

let string = array char ;;

let builtins = ref
  [ "+", repr prefix + , arr2 int int int ;
    "-", repr prefix - , arr2 int int int ;
    "*", repr prefix * , arr2 int int int ;
    "/", repr prefix / , arr2 int int int ;
    "mod", repr prefix mod , arr2 int int int ;
    "~", repr minus , arr int int ;
    "+.", repr prefix +. , arr2 float float float ;
    "-.", repr prefix -. , arr2 float float float ;
    "*.", repr prefix *. , arr2 float float float ;
    "/.", repr div_float , arr2 float float float ;
    "**", repr power, arr2 float float float ;
    "~.", repr minus_float , arr float float ;
    "=", repr prefix = , arr2 a a bool ;
    "<>", repr prefix <>, arr2 a a bool ;
    "==", repr prefix = , arr2 a a bool ;
    "!=", repr prefix = , arr2 a a bool ;
    "<", repr prefix < , arr2 a a bool ;
    ">", repr prefix > , arr2 a a bool ;
    "<=", repr prefix <= , arr2 a a bool ;
    ">=", repr prefix >= , arr2 a a bool ;
    "&", repr (fun a b -> a & b) , arr2 bool bool bool ;
    "or", repr (fun a b -> a or b) , arr2 bool bool bool ;
    "not", repr prefix not , arr bool bool ;
    "@", repr prefix @ , arr2 (list a) (list a) (list a) ;
    "^", repr concat_vect , arr2 (array a) (array a) (array a) ;
    ".", repr vect_item , arr2 (array a) int a ;
    ".<-", repr vect_assign , arr2 (array a) int (arr a unit);
    "$.", repr obj_field , unit ;
    "$.<-", repr set_obj_field , unit ;
(*    "get_field", repr vect_item ;
    "set_field", repr vect_assign ; *)
    "length", repr vect_length , arr (array a) int ;
    "array", repr make_vect , arr2 int a (array a) ;
    "copy", repr copy_vect , arr (array a) (array a) ;
    "sub", repr sub_vect , arr2 (array a) int (arr int (array a));
    "map_array", repr map_vect , arr2 (arr a b) (array a) (array b);
    "do_array", repr do_vect , arr2 (arr a b) (array a) unit ;
    "array_of_list", repr vect_of_list , arr (list a) (array a);
    "list_of_array", repr list_of_vect , arr (array a) (list a);
    "map_list", repr map , arr2 (arr a b) (list a) (list b);
    "do_list", repr do_list , arr2 (arr a b) (list a) unit ;
    "raise", repr minimal_fail , arr string a ;
    "print_string", repr (fun arr -> print_string (string_of_array arr)),
    arr string unit;
    "newline", repr print_newline, arr unit unit;
    "read_string", repr (fun () -> array_of_string (read_line ())),
    arr unit string;
    "string_of_int", repr (fun n -> array_of_string (string_of_int n)),
    arr int string;
    "string_of_float", repr(fun x -> array_of_string (string_of_float x)),
    arr float string;
    "print_int", repr print_int , arr int unit ;
    "print_float", repr print_float , arr float unit ;
    "read_int", repr read_int , arr unit int ;
    "read_float", repr read_float , arr unit float;
    "code", repr int_of_char, arr char int;
    "char", repr char_of_int, arr int char;
    "random_init", repr random__init, arr int unit;
    "random_int", repr random__int, arr int int;
    "random_float", repr random__float, arr float float;
    "maxint", repr maxint, int;
    "minint", repr minint, int;
    "float", repr float_of_int, arr int float;
    "trunc", repr truncate, arr float int;
    "exp", repr exp, arr float float;
    "log", repr log, arr float float;
    "sqrt", repr sqrt, arr float float;
    "sin", repr sin, arr float float;
    "cos", repr cos, arr float float;
    "tan", repr tan, arr float float;
    "asin", repr asin, arr float float;
    "acos", repr acos, arr float float;
    "atan", repr atan, arr float float;
    "abs", repr abs_float, arr float float;
    "trace", repr (trace maxint), arr string unit;
    "tracen", repr trace, arr2 int string unit;
    "untrace", repr untrace, arr string unit;
    "untrace_all", repr untrace_all, arr unit unit;
    "quit", repr (fun () -> sys__exit 0), arr unit unit
  ]
;;

let add_builtins () =
  do_list
    (fun (name, obj, ty) ->
      add_value (new_id name) obj
          {vi_type = ty; vi_access = Immutable})
    !builtins ;;

