(* $Id: toploop.ml,v 1.12 1997/04/18 10:46:32 garrigue Exp $ *)

#open"misc";;
#open"common";;
#open"syntax";;
#open"types";;
#open"define";;
#open"typechk";;
#open"compile";;
#open"printer";;
#open"builtins";;
#open"format";;
#open"lexing";;
#open"printf";;

let rec diff_list l1 l2 =
  if l1 == l2 then [] else hd l1 :: diff_list (tl l1) l2
;;

exception Line_too_long ;;

let feed_lexer ic lines buffer n =
  let add_line len =
    lines := len + hd !lines :: !lines in
  try
    let s = input_line ic in
    let len = string_length s in
    if len > n - 1 then raise Line_too_long;
    blit_string s 0 buffer 0 len;
    buffer.[len] <- `\n`;
    add_line (len + 1);
    len + 1
  with
    End_of_file -> 0
;;

let rec get_line n lines =
  match lines with
    [] -> (0,0)
  | start :: rest ->
      if start > n then get_line n rest else (list_length lines, start)
;;

let output_location filename lines oc loc =
  if filename <> "" then
    fprintf oc "In \"%s\", line " filename
  else fprintf oc "Line ";
  let (line,offset) = get_line loc.first lines in
  fprintf oc "%d, char %d-%d : "
    line (loc.first - offset) (loc.last - offset - 1)
;;

exception Toplevel ;;

let verbose = ref true;;

let do_phrase ph =
  try
    match ph with
      {sc_desc = SEexpr e} ->
	let ty = type_expression !values e in
	let ue = translate__expression !global_idents e in
	let obj = compile_expression ue () in
	flush stdout; flush stderr;
	if !verbose then begin
	  open_hovbox 1;
	  print_string "- : ";
	  print_scheme ty;
	  print_string " =";
	  print_space ();
	  print_value ty obj;
	  close_box ();
	  print_newline ()
	end
    | {sc_desc = STtype dl} ->
	add_typedef ph.sc_loc dl;
	if !verbose then
	  do_list
	    (fun td ->
	      print_endline ("type " ^ td.sd_name ^ " defined."))
	    dl
    | {sc_desc = SThide s} ->
	hide_type ph.sc_loc s;
	print_endline ("type " ^ s ^ " is now abstract.")
    | cmd ->
	let new_values = type_command !values cmd in
	let ucmds, new_idents = translate__command !global_idents cmd in
	let idents = diff_list new_idents !global_idents in
	let objs = compile_commands ucmds (map snd idents) in
	flush stdout; flush stderr;
	do_list2
	  (fun (s,id) obj -> add_value id obj (map__find s new_values))
	  idents objs;
	if !verbose then
	  do_list2
	    (fun (s,_) obj ->
	      let info = map__find s new_values in
	      open_hovbox 1;
	      print_string 
	      (if info.vi_access = Mutable then "var " else "val ");
	      print_string (s ^ " :");
	      print_space ();
	      print_scheme info.vi_type;
	      print_string " =";
	      print_space ();
	      print_value info.vi_type obj;
	      close_box ();
	      print_newline ())
	    idents objs
  with exn ->
    if handle exn then raise Toplevel else raise exn
;;

let report_error lexbuf filename lines exn =
  begin match exn with
    parsing__Parse_error ->
      eprintf "> %aSyntax error." (output_location filename lines)
        {first = get_lexeme_start lexbuf; last = get_lexeme_end lexbuf};
      prerr_endline ""
  | lexer__Lexical_error (_, start, stop) ->
      eprintf "> %aLexical error." (output_location filename lines)
        {first = start; last = stop};
      prerr_endline ""
  | Type_error (err, loc) ->
      set_formatter_out_channel stderr;
      eprintf "> %a" (output_location filename lines) loc;
      let (o,f) = get_formatter_output_functions () in
      report_error err;
      print_string "."; print_newline ();
      set_formatter_output_functions o f
  | Line_too_long ->
      eprintf "> Too long line in \"%s\", line %d."
        filename (list_length lines);
      prerr_endline ""
  | sys__Break ->
      prerr_endline "Interrupted!"
  | _ -> raise exn
  end;
  raise Toplevel
;;

let use s =
  let name = string_of_array s ^ ".mal" in
  let ic =
    try open_in name
    with sys__Sys_error _ ->
      prerr_endline ("> Could not open \"" ^ name ^ "\".");
      raise Toplevel
  in
  let lines = ref [0] in
  let lexbuf = create_lexer (feed_lexer ic lines) in
  try
    while true do do_phrase (parser__Phrase lexer__main lexbuf) done
  with exn ->
    close_in ic;
    match exn with
      End_of_file -> ()
    | _ -> report_error lexbuf name !lines exn
;;

builtins := ("use", obj__repr use, arr string unit) :: !builtins ;;

let minimal () =
  sys__catch_break true;
  open_hbox ();
  print_string ">\tMiniMAL version 0.03 on Caml Light";
  print_newline (); print_newline ();
  try while true do
    let lines = ref [0] in
    let lexbuf = create_lexer (feed_lexer stdin lines) in
    try
      while true do
	flush stdout; flush stderr;
	print_string "# ";
	print_flush ();
	lines := [hd !lines];
	do_phrase (parser__Phrase lexer__main lexbuf)
      done
    with exn ->
      try report_error lexbuf "" !lines exn
      with Toplevel -> ()
  done with End_of_file -> close_box ()
;;

add_builtins ();;

printexc__f minimal ();;
